"""Script to analyse NV center data from confocal microscope

Measurement conditions were not always the same, e.g. the laser power
changed, but data that is plotted on the same axis, was taken under
the same conditions.

"""

from datetime import datetime
from pathlib import Path

import h5py
import numpy as np
import matplotlib
import matplotlib.ticker
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d

import srim

data_path = Path(".")

matplotlib.rc("font", size="10")

#######################################################
#
# some helper functions to load and explore the data
#
#######################################################


def print_time(filename: Path):
    """Timezone is not clear, but this is still useful to make it easier to sort files by time"""
    unix_time = int(filename.name[:10])
    print(datetime.utcfromtimestamp(unix_time).strftime("%Y-%m-%d %H:%M:%S"))


def plot_2d_map(data, left=500, right=1500, extent=True):
    """Function used during data exploration"""
    if "map" in data:
        v = data["v_array"]
        h = data["h_array"]
        img = data["map"]
        e = data["extent"]
    else:

        w = data["wls"]
        v = data["v_array"]
        h = data["h_array"]
        d = data["spec_map"]
        e = data["extent"]

        mask = (w > left) & (w < right)
        img = d[:, :, mask].sum(axis=2)

    if extent is False:
        extent = None

    fig, ax = plt.subplots()
    ax.imshow(img, extent=e, origin="lower")
    ax.set_xlabel("X [um]")
    ax.set_xlabel("Z [um]")
    plt.show()


def load_data(filename, apd=False):
    """Files can be in two different format, with the apd flag, we can load both"""
    if apd:
        with h5py.File(filename, "r") as f:
            integration_time = f["hardware/ocean_optics_spectrometer/settings"].attrs[
                "oo_spec_int_time"
            ]
            print(filename, integration_time)
            data = {
                "v_array": f["apd_confocal"]["v_array"][()],
                "h_array": f["apd_confocal"]["h_array"][()],
                "map": f["apd_confocal"]["count_rate_map"][()] / integration_time,
                "extent": f["apd_confocal"].attrs["imshow_extent"],
            }
        return data

    with h5py.File(filename, "r") as f:
        integration_time = f["hardware/ocean_optics_spectrometer/settings"].attrs[
            "oo_spec_int_time"
        ]
        print(filename, integration_time)
        data = {
            "wls": f["spec_scan"]["wls"][()],
            "v_array": f["spec_scan"]["v_array"][()],
            "h_array": f["spec_scan"]["h_array"][()],
            "spec_map": f["spec_scan"]["spec_map"][()] / integration_time,
            "extent": f["spec_scan"].attrs["imshow_extent"],
        }
        return data


def pl(filename, apd=False):
    """shortcut during data exploration"""
    data = load_data(data_path / filename, apd=apd)
    plot_2d_map(data)


def ph(filename, apd=False):
    """shortcut during data exploration"""
    if not isinstance(filename, list):
        filename = [filename]
    fig, ax = plt.subplots()
    for f in filename:
        data = load_data(data_path / f, apd=apd)
        if "map" in data:
            ax.plot(data["map"].sum(axis=1))
        else:
            ax.plot(data["spec_map"].sum(axis=2).sum(axis=1))

    plt.show()


def list_file(hdf_file, output=True):
    """Lists all attributes and membersin an hdf5 files."""

    with h5py.File(data_path / hdf_file, "r") as f:
        content = {}

        def list_group(name):
            """callback function that will go through all items in the file"""
            nonlocal content

            if type(f[name]) == h5py._hl.dataset.Dataset:
                if output:
                    print(f"  {name} DATA {f[name].shape}")
                content[name] = f[name].shape
            else:
                # print attributes
                for k, v in f[name].attrs.items():
                    if output:
                        print(f"  {name} -> {k}: {v}")
                    content[f"{name}.{k}"] = v
                # print other times, but skip data (already handled)
                for k, v in f[name].items():
                    if type(v) == h5py._hl.dataset.Dataset:
                        continue
                    if type(v) == h5py._hl.group.Group:
                        continue
                    print(f"TODO add message for {type(v)}")

        f.visit(list_group)
        return content


#######################################################
#
#  The data files
#
#######################################################


# SHI irradiated data from folder 20160331_session
# Taken at laser powerwheel setting of 1800

SHI_irradiated_file = data_path / "1459459109_spec_scan.h5"  # 2016-03-31 21:18:29
SHI_irradiated_data = load_data(SHI_irradiated_file)

SHI_non_irradiated_file = data_path / "1459460392_spec_scan.h5"  # 2016-03-31 21:39:52
SHI_non_irradiated_data = load_data(SHI_non_irradiated_file)

# measurement of laser reflection to figure out Z=0
# data_path / "1459461809_spec_scan.h5"  # depth alignment  2016-03-31 22:03:29
# data_path / "1459461962_spec_scan.h5"  # DA               2016-03-31 22:06:02

# post-aneal laser reflection measurement '20160712_session/160711'

# reflection data
# data_path / "1468264790_apd_confocal.h5"  # DA  all 4 have peaks   2016-07-11 19:19:50 light area
# data_path / "1468264843_apd_confocal.h5"  # DA                     2016-07-11 19:20:43 dark area
# data_path / "1468264919_apd_confocal.h5"  # DA                     2016-07-11 19:21:59 before measurement
# data_path / "1468271729_apd_confocal.h5"  # DA                     2016-07-11 21:15:29 after measurement


# SHI_and_annealed_file "20160712_session/160711"
# Powerwheel (= laser power) @ 1000 <- lower power level than SHI irradiated

# XZ scan, DARK h_array[0:20].sum, LIGHT h_array(119:133)    2016-07-11 19:56:51
SHI_irradiated_and_annealed_file = data_path / "1468267011_spec_scan.h5"
SHI_irradiated_and_annealed_data = load_data(SHI_irradiated_and_annealed_file)


# XY overview shown in Fig 1.
overviewXY = data_path / "1468271221_apd_confocal.h5"  #  2016-07-11 21:07:01
overviewXY_data = load_data(overviewXY, apd=True)


# reference sample 1ppm  20160712_session/160711": currently not used
control_file = data_path / "1468273317_spec_scan.h5"
control_data = load_data(control_file)

control_surface = data_path / "1468273191_apd_confocal.h5"

#######################################################
#
# Figure 1
#
#######################################################

fig, ax = plt.subplots(
    2,
    2,
    figsize=(8, 4),
    constrained_layout=True,
    gridspec_kw={"width_ratios": [1.5, 1]},
)

# we zoom into the correct X and Z location to only get spectra that show NV centers for better signal to noise
d = SHI_irradiated_data["spec_map"]
w = SHI_irradiated_data["wls"]
zoom_irradiated = d[70:82, :, :]
zoom_irradiated = zoom_irradiated / zoom_irradiated.shape[0] / zoom_irradiated.shape[1]
ax[0][0].plot(
    w, zoom_irradiated.sum(axis=0).sum(axis=0), color="red", label="irradiated"
)

d = SHI_non_irradiated_data["spec_map"]
w = SHI_non_irradiated_data["wls"]
zoom_non_irradiated = d[70:82, :, :]
zoom_non_irradiated = (
    zoom_non_irradiated / zoom_non_irradiated.shape[0] / zoom_non_irradiated.shape[1]
)
ax[0][0].plot(
    w,
    zoom_non_irradiated.sum(axis=0).sum(axis=0),
    "-.",
    color="grey",
    label="non-irradiated",
)
ax[0][0].set_ylabel("Intensity [a.u.]")
ax[0][0].set_title("SHI Irradiated")

photo = plt.imread("photo.jpg", format="jpg")
ax[0][1].set_title("SHI Irradiated Sample")
ax[0][1].imshow(photo)
ax[0][1].axis("off")

d = SHI_irradiated_and_annealed_data["spec_map"]
w = SHI_irradiated_and_annealed_data["wls"]

# average region to reduce noise use same depth slice, but different X ranges for inside/outside of irradiated area
zoom_irradiated = d[96:111, :21, :]
zoom_irradiated = zoom_irradiated / zoom_irradiated.shape[0] / zoom_irradiated.shape[1]
ax[1][0].plot(
    w, zoom_irradiated.sum(axis=0).sum(axis=0), color="red", label="irradiated (i)"
)

zoom_non_irradiated = d[96:111, 119:, :]
zoom_non_irradiated = (
    zoom_non_irradiated / zoom_non_irradiated.shape[0] / zoom_non_irradiated.shape[1]
)
ax[1][0].plot(
    w,
    zoom_non_irradiated.sum(axis=0).sum(axis=0),
    "-.",
    color="grey",
    label="non-irradiated (ii)",
)
ax[1][0].set_xlabel("Wavelength [nm]")
ax[1][0].set_ylabel("Intensity [a.u.]")
ax[1][0].set_title("SHI Irradiated + Annealed")
ax[1][0].legend(loc="best")

ax[0][0].set_yscale("log")
ax[0][0].yaxis.set_minor_locator(
    matplotlib.ticker.LogLocator(subs=np.arange(2, 10) * 0.1, numticks=14)
)
ax[0][0].set_ylim([1e2, 1e5])
ax[1][0].set_yscale("log")
ax[1][0].yaxis.set_minor_locator(
    matplotlib.ticker.LogLocator(subs=np.arange(2, 10) * 0.1, numticks=14)
)
ax[1][0].set_ylim([1e1, 1e4])

img = overviewXY_data["map"]
e = overviewXY_data["extent"]
im = ax[1][1].imshow(img / img.max(), extent=e, origin="lower", aspect="auto")
plt.colorbar(im)
ax[1][1].set_xlabel("X [$\mu$m]")
ax[1][1].set_ylabel("Y [$\mu$m]")
ax[1][1].set_title("Top View Intensity Map")

fig.text(0.02, 0.9, "a)", fontsize=17)
fig.text(0.02, 0.475, "b)", fontsize=17)
fig.text(0.55, 0.9, "c)", fontsize=17)
fig.text(0.55, 0.475, "d)", fontsize=17)

fig.text(0.67, 0.25, "(i)", fontsize=17, bbox=dict(facecolor="white", alpha=0.5))
fig.text(0.82, 0.25, "(ii)", fontsize=17, bbox=dict(facecolor="white", alpha=0.5))

plt.savefig("Fig1.pdf")
plt.show()

#######################################################
#
# Code for Figure 3 in the paper
#
#######################################################

# find the surface
fig, ax = plt.subplots()
for i, d in enumerate(
    [
        data_path / "1459461809_spec_scan.h5",
        data_path / "1459461962_spec_scan.h5",
        data_path / "1468264919_apd_confocal.h5",
        data_path / "1468271729_apd_confocal.h5",
        control_surface,
    ]
):
    apd = "apd" in d.name
    data = load_data(d, apd=apd)
    key = "map" if apd else "spec_map"
    z = data["v_array"]
    if apd:
        data = data[key][:, :].sum(axis=1)
        data = data / data.max()
        plt.plot(z, data, "-o", label=f"{i}")
    else:
        data = data[key][:, :, 50:100].sum(axis=1).sum(axis=1)
        data = data / data.max()
        plt.plot(z, data, "-o", label=f"{i}")
plt.legend()
plt.show()

# read of Z location (no need to fit, since error will be small < 0.3 um and we later plot on a scale of 60 um)
# 0: 25.4 pre
# 1: 25.5 pre
# 2: 34.1 post before scan
# 3: 34.9 post after scan
# use this later:
z_origin = {"pre": 25.45, "post": 34.5, "control": 27.5}

# zoom in on the NV peak, i.e. find integration boundaries
# this plot is not published
fig, ax = plt.subplots(nrows=1, ncols=1)

d = SHI_irradiated_data["spec_map"]
d = d.sum(axis=1).sum(axis=0)
ax.plot(d, color="red", label="irradiated")

d = SHI_non_irradiated_data["spec_map"]
d = d.sum(axis=1).sum(axis=0)
ax.plot(d, "-.", color="grey", label="non-irradiated")

d = SHI_irradiated_and_annealed_data["spec_map"]
d = d[96:111, :21, :].sum(axis=1).sum(axis=0)
ax.plot(d, color="red", label="irradiated")

d = SHI_irradiated_and_annealed_data["spec_map"]
d = d[96:111, 119:, :].sum(axis=1).sum(axis=0)
ax.plot(d, "-.", color="grey", label="non-irradiated")
ax.legend(loc="best")
ax.set_title("Test for finding integration for NV peak")
NV_left = 180
NV_right = 225
ax.axvline(NV_left)
ax.axvline(NV_right)
plt.show()

# final plot for paper

# also tried to do a fit to the data with Linear+Gaussian
# model, but didn't get much better results, so I kept the simpler
# version of just integrating across the 637 nm line

fig, ax = plt.subplots(
    nrows=4,
    ncols=1,
    sharex=True,
    figsize=(6, 7),
    # gridspec_kw={"hspace": 0.2},
    constrained_layout=True,
)

# data = np.loadtxt(data_path / "IONIZ.txt", skiprows=28)
# electronic_stopping_Z = data[:, 0] / 10 / 1000  # Angstrom to um
# electronic_stopping = (data[:, 1] + data[:, 2]) * 10  # eV/nm
# data = np.loadtxt(data_path / "E2RECOIL.txt", skiprows=27)
# nuclear_stopping_Z = data[:, 0] / 10 / 1000  # Angstrom to um
# nuclear_stopping = data[:, 1] * 10  # vacancies/nm

data = srim.output.SRResults(".")
energy = data.data[0]  # keV
el_stop = data.data[1] * 10  # eV/nm
nu_stop = data.data[2] * 10  # eV/nm

# add zero energy value
energy = np.insert(energy, 0, 0)
el_stop = np.insert(el_stop, 0, 0)
nu_stop = np.insert(nu_stop, 0, 0)

# create splines
el_stopping = interp1d(energy, el_stop, kind="cubic")
nu_stopping = interp1d(energy, nu_stop, kind="cubic")

# create energy vs depth by taking 1 nm steps and tracking the energy
E = 1.1e6  # ub keV
loc = 0  # in um
depth = [loc]
energy = [E]
while E > 1:
    loss = el_stopping(E) + nu_stopping(E)
    loss *= 1e-3
    loc += 1 * 1e-3  # 1 nm in um
    E -= loss
    depth.append(loc)
    energy.append(E)
# interploate um depth vs energy in keV
E = interp1d(depth, energy, kind="cubic", fill_value=0, bounds_error=False)
pos = np.linspace(0, 40, 81)  # in um
energy = E(pos)  # energies at depth

# now look up the stopping power at that depth
electronic_stopping = el_stopping(energy)
nuclear_stopping = nu_stopping(energy)
electronic_stopping_Z = pos
nuclear_stopping_Z = pos

es = ax[0].plot(
    electronic_stopping_Z,
    electronic_stopping,
    color="blue",
    label="electronic stopping",
)
ax[0].set_ylabel("Electronic Stopping\n[eV/nm]")
right_axis = ax[0].twinx()
ns = right_axis.plot(
    nuclear_stopping_Z, nuclear_stopping, "-.", color="green", label="nuclear stopping"
)
right_axis.set_ylabel("Nuclear Stopping\n[eV/nm]")

ax[0].legend(es + ns, ["electronic\nstopping", "nuclear\nstopping"], loc="lower right")
ax[0].set_title("SRIM Simulation")


def DR(z, NA=0.95, n=2.4):
    """Equation 1 from Everall"""
    return z * (np.sqrt((n * n - 1) * NA * NA / (1 - NA * NA) + n * n) - n)


n = 2.4
NA = 0.95
right_axis2 = ax[1].twinx()

y = np.zeros_like(electronic_stopping_Z)
for i, z in enumerate(electronic_stopping_Z):
    z_at_air = z / n
    dr = DR(z_at_air)
    mask = (electronic_stopping_Z > z) & (electronic_stopping_Z < z + dr)
    y[i] = electronic_stopping[mask].sum() / mask.sum()
ax[1].plot(electronic_stopping_Z, y, color="blue", label=f"electronic\nstopping")
ax[1].set_ylabel("Electronic Stopping\n[a.u.]")

y = np.zeros_like(nuclear_stopping_Z)
for i, z in enumerate(nuclear_stopping_Z):
    z_at_air = z / n
    dr = DR(z_at_air, NA=NA)
    mask = (nuclear_stopping_Z > z) & (nuclear_stopping_Z < z + dr)
    y[i] = nuclear_stopping[mask].sum() / mask.sum()
right_axis2.plot(nuclear_stopping_Z, y, "-.", color="green", label=f"nuclear\nstopping")
right_axis2.set_ylabel("Nuclear Stopping\n[a.u.]")
ax[1].set_title("SRIM Simulation & Optical Readout")


d = SHI_irradiated_data["spec_map"]
z = SHI_irradiated_data["v_array"]
z = z_origin["pre"] - z
z[z > 0] = z[z > 0] * 2.4
d = d[:, :, NV_left:NV_right].sum(axis=2).sum(axis=1)
mymax = d.max()
ax[2].plot(z, d / mymax, color="red", label="irradiated")

d = SHI_non_irradiated_data["spec_map"]
z = SHI_non_irradiated_data["v_array"]
z = z_origin["pre"] - z
z[z > 0] = z[z > 0] * 2.4
d = d[:, :, NV_left:NV_right].sum(axis=2).sum(axis=1)
ax[2].plot(z, d / mymax, "-.", color="grey", label="non-irradiated")

d = SHI_irradiated_and_annealed_data["spec_map"]
z = SHI_irradiated_and_annealed_data["v_array"]
z = z_origin["post"] - z
z[z > 0] = z[z > 0] * 2.4
d = d[:, :21, NV_left:NV_right].sum(axis=2).sum(axis=1)
mymax = d.max()
ax[3].plot(z, d / mymax, color="red", label="irradiated")

d = SHI_irradiated_and_annealed_data["spec_map"]
z = SHI_irradiated_and_annealed_data["v_array"]
z = z_origin["post"] - z
z[z > 0] = z[z > 0] * 2.4
d = d[:, 119:, NV_left:NV_right].sum(axis=2).sum(axis=1)
ax[3].plot(z, d / mymax, "-.", color="grey", label="non-irradiated")

d = control_data["spec_map"]
z = control_data["v_array"]
z = z_origin["control"] - z
z[z > 0] = z[z > 0] * 2.4
d = d[:, :, NV_left:NV_right].sum(axis=2).sum(axis=1)
ax[3].plot(z, d / mymax, "--", color="black", label="control")

ax[0].set_yscale("log")
ax[0].yaxis.set_minor_locator(matplotlib.ticker.LogLocator(subs="all"))
ax[0].set_ylim([2e3, 2e5])
right_axis.set_yscale("log")
right_axis.yaxis.set_minor_locator(
    matplotlib.ticker.LogLocator(subs=np.arange(2, 10) * 0.1, numticks=14)
)
right_axis.set_ylim([0.7e1, 5e4])
ax[1].set_yscale("log")
ax[1].yaxis.set_minor_locator(matplotlib.ticker.LogLocator(subs="all"))
ax[1].set_ylim([2e3, 2e5])
right_axis2.set_yscale("log")
right_axis2.yaxis.set_minor_locator(
    matplotlib.ticker.LogLocator(subs=np.arange(2, 10) * 0.1, numticks=14)
)
right_axis2.set_ylim([0.7e1, 5e4])
ax[2].set_yscale("log")
ax[2].yaxis.set_minor_locator(matplotlib.ticker.LogLocator(subs="all"))
ax[3].set_yscale("log")
ax[3].yaxis.set_minor_locator(matplotlib.ticker.LogLocator(subs="all"))
ax[3].set_ylim([0.5e-2, 2])

ax[3].set_xlabel("Z [$\mu$m]")

ax[2].set_title("SHI Irradiated")
ax[3].set_title("SHI Irradiated + Annealed")

ax[2].set_ylabel("Counts [a.u.]")
ax[3].set_ylabel("Counts [a.u.]")

for a in ax:
    a.set_xlim([0, 55])

ax[3].legend(loc="upper right")

fig.text(0.022, 0.97, "a)", fontsize=17)
fig.text(0.022, 0.75, "b)", fontsize=17)
fig.text(0.022, 0.50, "c)", fontsize=17)
fig.text(0.022, 0.25, "d)", fontsize=17)

plt.savefig("Fig3.pdf")
plt.show()
